# microservice_manager.py

from typing import Dict, List, Set

from microservice_story_manager import UserStoryManager, UserStory
from microservice_story_manager.error_messages import get_error_message


class Microservice:
    def __init__(self, identifier: str):
        self.identifier = identifier
        self.stories: Set[str] = set()


class MicroserviceManager:
    def __init__(self, user_story_manager: UserStoryManager):
        self.microservices: Dict[str, Microservice] = {}
        self.user_story_manager = user_story_manager
        self.project_language: str = 'en'  # Default language
        self.next_id: int = 1
        self.story_counter = {}

    def set_project_language(self, language: str) -> None:
        self.project_language = language

    def create_microservice(self) -> str:
        identifier = f"ms{self.next_id:03d}"
        self.microservices[identifier] = Microservice(identifier)
        self.next_id += 1
        return identifier

    def add_story_to_microservice(self, microservice_id: str, story_id: str) -> None:
        if microservice_id not in self.microservices:
            raise ValueError(f"Microservice with ID {microservice_id} does not exist.")

        story = self.user_story_manager.get_story(story_id)
        if story is None:
            raise ValueError(f"Story with ID {story_id} does not exist.")

        self.microservices[microservice_id].stories.add(story_id)
        """print(f"Added story {story_id} to microservice {microservice_id}")
        print(f"Story dependencies: {story.dependencies}")"""

    def remove_story_from_microservice(self, microservice_id: str, story_id: str) -> None:
        if microservice_id not in self.microservices:
            raise ValueError(get_error_message('story_not_found', self.project_language).format(microservice_id))

        if story_id not in self.microservices[microservice_id].stories:
            raise ValueError(get_error_message('story_not_found', self.project_language).format(story_id))

        self.microservices[microservice_id].stories.remove(story_id)

        if not self.microservices[microservice_id].stories:
            raise ValueError(get_error_message('microservice_empty', self.project_language))

    def get_microservice_stories(self, microservice_id: str) -> List[UserStory]:
        if microservice_id not in self.microservices:
            raise ValueError(get_error_message('story_not_found', self.project_language).format(microservice_id))

        return [self.user_story_manager.get_story(story_id) for story_id in self.microservices[microservice_id].stories]

    def get_all_microservices(self) -> Dict[str, List[UserStory]]:
        return {ms_id: self.get_microservice_stories(ms_id) for ms_id in self.microservices}

    def validate_all_stories_assigned(self) -> bool:
        all_stories = set(self.user_story_manager.get_all_stories().keys())
        assigned_stories = set()
        for ms in self.microservices.values():
            assigned_stories.update(ms.stories)

        unassigned_stories = all_stories - assigned_stories
        if unassigned_stories:
            raise ValueError(
                f"The following stories are not assigned to any microservice: {', '.join(unassigned_stories)}")
        return True

    def get_microservice_dependencies(self) -> Dict[str, Set[str]]:
        dependencies = {ms_id: set() for ms_id in self.microservices}
        for ms_id, ms in self.microservices.items():
            for story_id in ms.stories:
                story = self.user_story_manager.get_story(story_id)
                for dep_story_id in story.dependencies:
                    for dep_ms_id, dep_ms in self.microservices.items():
                        if dep_story_id in dep_ms.stories and dep_ms_id != ms_id:
                            dependencies[ms_id].add(dep_ms_id)
        return dependencies





import csv
from typing import Dict, List, Set, Union

import matplotlib.pyplot as plt
import networkx as nx

from microservice_story_manager.error_messages import get_error_message


class UserStory:
    def __init__(self, title: str, description: str, language: str, dependencies: List[str],
                 story_points: int, estimated_time: float, scenario: str, observations: str):
        self.identifier = None
        self.title = title
        self.description = description
        self.language = language
        self.dependencies = dependencies
        self.story_points = story_points
        self.estimated_time = estimated_time
        self.scenario = scenario
        self.observations = observations


class UserStoryManager:
    def __init__(self):
        self.stories: Dict[str, UserStory] = {}
        self.pending_dependencies: Dict[str, Set[str]] = {}
        self.project_language: str = 'en'  # Default language

    def set_project_language(self, language: str) -> None:
        self.project_language = language

    def add_story(self, identifier: str, title: str, description: str, language: str,
                  dependencies: List[str], story_points: Union[int, str],
                  estimated_time: Union[float, str], scenario: str, observations: str) -> None:
        if identifier in self.stories:
            raise ValueError(get_error_message('story_exists', self.project_language).format(identifier))

        try:
            story_points = int(story_points)
            if story_points <= 0:
                raise ValueError
        except ValueError:
            raise ValueError(get_error_message('invalid_story_points', self.project_language))

        try:
            estimated_time = float(estimated_time)
            if estimated_time <= 0:
                raise ValueError
        except ValueError:
            raise ValueError(get_error_message('invalid_estimated_time', self.project_language))

        self.stories[identifier] = UserStory(title, description, language, dependencies,
                                             story_points, estimated_time, scenario, observations)

        if identifier in self.pending_dependencies:
            del self.pending_dependencies[identifier]

        for dep in dependencies:
            if dep not in self.stories:
                if dep not in self.pending_dependencies:
                    self.pending_dependencies[dep] = set()
                self.pending_dependencies[dep].add(identifier)

    def get_story(self, identifier: str) -> UserStory:
        if identifier not in self.stories:
            raise ValueError(get_error_message('story_not_found', self.project_language).format(identifier))
        return self.stories[identifier]

    def get_all_stories(self) -> Dict[str, UserStory]:
        return self.stories

    def validate_integrity(self) -> bool:
        if self.pending_dependencies:
            missing = ", ".join(self.pending_dependencies.keys())
            raise ValueError(get_error_message('missing_dependencies', self.project_language).format(missing))
        return True

    def get_total_story_points(self) -> int:
        return sum(story.story_points for story in self.stories.values())

    def get_total_estimated_time(self) -> float:
        return sum(story.estimated_time for story in self.stories.values())

    def plot_stories(self) -> None:
        G = nx.DiGraph()

        for identifier, story in self.stories.items():
            G.add_node(identifier,
                       story_points=story.story_points,
                       title=story.title)
            for dep in story.dependencies:
                G.add_edge(dep, identifier)

        pos = nx.spring_layout(G)

        plt.figure(figsize=(12, 8))

        node_sizes = [story.story_points * 100 for story in self.stories.values()]

        nx.draw(G, pos, with_labels=False, node_color='lightblue',
                node_size=node_sizes, arrows=True)

        labels = {identifier: f"{story.title}\n({story.story_points} SP)" for identifier, story in self.stories.items()}
        nx.draw_networkx_labels(G, pos, labels, font_size=8)

        plt.title("User Stories Dependency Graph")
        plt.axis('off')
        plt.tight_layout()
        plt.show()

    def load_stories_from_csv(self, file_path: str) -> None:
        with open(file_path, 'r', newline='', encoding='utf-8') as file:
            csv_reader = csv.DictReader(file)
            for row in csv_reader:
                try:
                    dependencies = [dep.strip() for dep in row['dependencies'].split(',') if dep.strip()]
                    self.add_story(
                        identifier=row['identifier'],
                        title=row['title'],
                        description=row['description'],
                        language=row['language'],
                        dependencies=dependencies,
                        story_points=row['story_points'],
                        estimated_time=row['estimated_time'],
                        scenario=row['scenario'],
                        observations=row['observations']
                    )
                except Exception as e:
                    print(f"Error adding story {row['identifier']}: {str(e)}")

        try:
            self.validate_integrity()
            print("All stories were loaded correctly and integrity is valid.")
        except ValueError as e:
            print(f"Integrity error: {str(e)}")

    def update_story_identifier(self, old_identifier: str, new_identifier: str) -> None:
        if old_identifier not in self.stories:
            raise ValueError(f"Story with ID {old_identifier} does not exist.")

        story = self.stories[old_identifier]
        story.identifier = new_identifier
        self.stories[new_identifier] = story
        del self.stories[old_identifier]



from microservice_story_manager import UserStoryManager, MicroserviceManager, MicroserviceVisualizer, \
    GeneticMicroserviceAssigner, FitnessFunctions, MetricsVisualizer, MicroserviceMetricsManager

# Initialize managers
user_story_manager = UserStoryManager()
user_story_manager.load_stories_from_csv('../examples/user_stories.csv')

microservice_manager = MicroserviceManager(user_story_manager)

# Create microservices
for _ in range(5):
    microservice_manager.create_microservice()

# Create FitnessFunctions instance
fitness_functions = FitnessFunctions(MicroserviceMetricsManager(microservice_manager))

# Choose fitness functions to compare
functions_to_compare = [
    (fitness_functions.f1, "f1", True),
    (fitness_functions.f2, "f2", True)
]

results = []

for func, name, minimize in functions_to_compare:
    assigner = GeneticMicroserviceAssigner(user_story_manager, microservice_manager, func, minimize)
    optimized_manager, fitness_value, detailed_metrics, story_assignments = assigner.optimize_assignments(generations=3, population_size=10)
    results.append((name, optimized_manager, fitness_value, detailed_metrics, story_assignments))

    print(f"\nResults for fitness function {name}:")
    print(f"Best fitness: {fitness_value}")
    print(f"Detailed metrics: {detailed_metrics}")
    print("\nStory Assignments:")
    for ms_id, stories in story_assignments.items():
        print(f"\nMicroservice {ms_id}:")
        for story in stories:
            print(f"  - {story}")

# Compare metrics
MetricsVisualizer.compare_metrics(results[0][3], results[1][3], results[0][0], results[1][0])

# Compare fitness values
MetricsVisualizer.compare_fitness_values([result[2] for result in results], [result[0] for result in results])

# Visualize the optimized microservices
for name, manager, _, _, _ in results:
    visualizer = MicroserviceVisualizer(manager)
    visualizer.plot_microservices(title=f"Optimized Microservices using {name}")



import math
from typing import Dict, List, Set, Tuple

from microservice_story_manager import MicroserviceManager, SemanticSimilarityCalculator


class MicroserviceMetricsManager:
    def __init__(self, microservice_manager: MicroserviceManager):
        self.microservice_manager = microservice_manager
        self.semantic_calculator = SemanticSimilarityCalculator(microservice_manager.user_story_manager)

    def _initialize_semantic_similarity(self):
        stories = []
        print("Initializing Semantic Similarity:")
        for ms_id, ms in self.microservice_manager.microservices.items():
            print(f"Processing Microservice {ms_id}")
            for story_id in ms.stories:
                story = self.microservice_manager.user_story_manager.get_story(story_id)
                if story.identifier:
                    stories.append((story.identifier, story.title, story.description))
                    print(f"  Added story: {story.identifier}")
                else:
                    print(f"  Skipped story with None identifier: {story.title}")

        print(f"Total stories collected: {len(stories)}")
        self.semantic_calculator.build_similarity_dict(stories)
        print("Finished building similarity dictionary")

    def calculate_ais_metric(self) -> Dict[str, int]:
        ais_metrics = {ms_id: 0 for ms_id in self.microservice_manager.microservices}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            dependent_microservices = set()
            for other_ms_id, other_ms in self.microservice_manager.microservices.items():
                if other_ms_id != ms_id:
                    other_stories = self.microservice_manager.get_microservice_stories(other_ms_id)
                    for story in other_stories:
                        if any(dep in microservice.stories for dep in story.dependencies):
                            dependent_microservices.add(other_ms_id)
                            break
            ais_metrics[ms_id] = len(dependent_microservices)
        return ais_metrics

    def calculate_ads_metric(self) -> Dict[str, int]:
        ads_metrics = {ms_id: 0 for ms_id in self.microservice_manager.microservices}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            dependent_on_microservices = set()
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            for story in stories:
                for dep in story.dependencies:
                    for other_ms_id, other_ms in self.microservice_manager.microservices.items():
                        if other_ms_id != ms_id and dep in other_ms.stories:
                            dependent_on_microservices.add(other_ms_id)
            ads_metrics[ms_id] = len(dependent_on_microservices)
        return ads_metrics

    def calculate_siy_metric(self) -> Dict[str, int]:
        """
        Calculate the Microservice Interdependence (SIY) metric for each microservice.
        SIY is the number of other microservices that are interdependent with a given microservice.
        """
        siy_metrics = {ms_id: 0 for ms_id in self.microservice_manager.microservices}
        ais_metrics = self.calculate_ais_metric()
        ads_metrics = self.calculate_ads_metric()

        for ms_id in self.microservice_manager.microservices:
            interdependent_count = 0
            for other_ms_id in self.microservice_manager.microservices:
                if ms_id != other_ms_id:
                    if ais_metrics[ms_id] > 0 and ads_metrics[other_ms_id] > 0 and \
                            ais_metrics[other_ms_id] > 0 and ads_metrics[ms_id] > 0:
                        interdependent_count += 1
            siy_metrics[ms_id] = interdependent_count

        return siy_metrics

    def calculate_ais_total(self) -> float:
        """
        Calculate the total AIS value at the system level (AisT).
        """
        ais_values = list(self.calculate_ais_metric().values())
        return math.sqrt(sum(ais ** 2 for ais in ais_values))

    def calculate_ads_total(self) -> float:
        """
        Calculate the total ADS value at the system level (AdsT).
        """
        ads_values = list(self.calculate_ads_metric().values())
        return math.sqrt(sum(ads ** 2 for ads in ads_values))

    def calculate_siy_total(self) -> float:
        """
        Calculate the total SIY value at the system level (SiyT).
        """
        siy_values = list(self.calculate_siy_metric().values())
        return math.sqrt(sum(siy ** 2 for siy in siy_values))

    def get_ais_vector(self) -> List[int]:
        """
        Get the AIS vector containing the AIS value for each microservice.
        """
        ais_metrics = self.calculate_ais_metric()
        return [ais_metrics[ms_id] for ms_id in sorted(ais_metrics.keys())]

    def get_ads_vector(self) -> List[int]:
        """
        Get the ADS vector containing the ADS value for each microservice.
        """
        ads_metrics = self.calculate_ads_metric()
        return [ads_metrics[ms_id] for ms_id in sorted(ads_metrics.keys())]

    def get_siy_vector(self) -> List[int]:
        """
        Get the SIY vector containing the SIY value for each microservice.
        """
        siy_metrics = self.calculate_siy_metric()
        return [siy_metrics[ms_id] for ms_id in sorted(siy_metrics.keys())]

    def calculate_coupling_total(self) -> float:
        ais_total, ads_total, siy_total = self.get_coupling_indicators()
        cpt = 10 * math.sqrt(ais_total ** 2 + ads_total ** 2 + siy_total ** 2)
        return cpt

    def get_coupling_indicators(self) -> tuple[float, float, float]:
        ais_total = self.calculate_ais_total()
        ads_total = self.calculate_ads_total()
        siy_total = self.calculate_siy_total()

        return ais_total, ads_total, siy_total

    def calculate_lc_metric(self) -> Dict[str, int]:
        """
        Calculate the Lack of Cohesion (LC) metric for each microservice.
        LC is the number of pairs of microservices not having any interdependency with the given microservice.
        """
        lc_metrics = {}
        total_microservices = len(self.microservice_manager.microservices)

        for ms_id in self.microservice_manager.microservices:
            dependent_microservices = set()
            for other_ms_id, other_ms in self.microservice_manager.microservices.items():
                if ms_id != other_ms_id:
                    if self._has_interdependency(ms_id, other_ms_id):
                        dependent_microservices.add(other_ms_id)

            lc_metrics[ms_id] = total_microservices - 1 - len(dependent_microservices)

        return lc_metrics

    def _has_interdependency(self, ms_id1: str, ms_id2: str) -> bool:
        """
        Check if there's an interdependency between two microservices.
        """
        ms1_stories = self.microservice_manager.get_microservice_stories(ms_id1)
        ms2_stories = self.microservice_manager.get_microservice_stories(ms_id2)

        for story in ms1_stories:
            if any(dep in self.microservice_manager.microservices[ms_id2].stories for dep in story.dependencies):
                return True

        for story in ms2_stories:
            if any(dep in self.microservice_manager.microservices[ms_id1].stories for dep in story.dependencies):
                return True

        return False

    def calculate_coh_metric(self) -> Dict[str, float]:
        """
        Calculate the Lack of Cohesion Grade (Coh) metric for each microservice.
        Coh is the proportion of LC divided by the total number of microservices.
        """
        lc_metrics = self.calculate_lc_metric()
        total_microservices = len(self.microservice_manager.microservices)

        coh_metrics = {ms_id: lc / total_microservices for ms_id, lc in lc_metrics.items()}
        return coh_metrics

    def calculate_coh_total(self) -> float:
        """
        Calculate the total Coh value at the system level (CohT).
        """
        coh_values = list(self.calculate_coh_metric().values())
        return math.sqrt(sum(coh ** 2 for coh in coh_values))

    def get_coh_vector(self) -> List[float]:
        """
        Get the Coh vector containing the Coh value for each microservice.
        """
        coh_metrics = self.calculate_coh_metric()
        return [coh_metrics[ms_id] for ms_id in sorted(coh_metrics.keys())]

    def get_cohesion_metrics(self) -> Tuple[Dict[str, int], Dict[str, float], float]:
        """
        Get all cohesion metrics: LC, Coh, and CohT.
        """
        lc_metrics = self.calculate_lc_metric()
        coh_metrics = self.calculate_coh_metric()
        coh_total = self.calculate_coh_total()

        return lc_metrics, coh_metrics, coh_total

    def calculate_number_of_microservices(self) -> int:
        """
        Calculate the Number of Microservices (N) in the system.
        """
        return len(self.microservice_manager.microservices)

    def calculate_wsic_metric(self) -> Dict[str, int]:
        """
        Calculate the Weighted Service Interface Count (WSIC) for each microservice.
        WSIC is the number of user stories assigned to each microservice.
        """
        wsic_metrics = {}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            wsic_metrics[ms_id] = len(microservice.stories)
        return wsic_metrics

    def calculate_wsic_total(self) -> int:
        """
        Calculate the total WSIC value at the system level (WsicT).
        WsicT is the maximum number of user stories associated with any microservice.
        """
        wsic_values = self.calculate_wsic_metric().values()
        return max(wsic_values) if wsic_values else 0

    def get_granularity_metrics(self) -> Tuple[int, Dict[str, int], int]:
        """
        Get all granularity metrics: N, WSIC, and WsicT.
        """
        n = self.calculate_number_of_microservices()
        wsic_metrics = self.calculate_wsic_metric()
        wsic_total = self.calculate_wsic_total()

        return n, wsic_metrics, wsic_total

    def calculate_calls_metric(self) -> Dict[str, int]:
        """
        Calculate the number of calls (Calls) for each microservice.
        Calls corresponds to the number of invocations of MSi to other microservices.
        """
        calls_metrics = {ms_id: 0 for ms_id in self.microservice_manager.microservices}

        for ms_id, microservice in self.microservice_manager.microservices.items():
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            for story in stories:
                for dep in story.dependencies:
                    if dep not in microservice.stories:
                        calls_metrics[ms_id] += 1

        return calls_metrics

    def calculate_requests_metric(self) -> Dict[str, int]:
        """
        Calculate the number of requests (Requests) for each microservice.
        Requests corresponds to the number of invocations of other microservices to MSi.
        """
        requests_metrics = {ms_id: 0 for ms_id in self.microservice_manager.microservices}

        for ms_id, microservice in self.microservice_manager.microservices.items():
            for other_ms_id, other_microservice in self.microservice_manager.microservices.items():
                if ms_id != other_ms_id:
                    other_stories = self.microservice_manager.get_microservice_stories(other_ms_id)
                    for story in other_stories:
                        if any(dep in microservice.stories for dep in story.dependencies):
                            requests_metrics[ms_id] += 1

        return requests_metrics

    def calculate_average_calls(self) -> float:
        """
        Calculate the Average of calls of MSBA (Avg. Calls).
        Avg. Calls is the average of calls among microservices of MSBA.
        """
        calls_metrics = self.calculate_calls_metric()
        total_calls = sum(calls_metrics.values())
        number_of_microservices = len(self.microservice_manager.microservices)

        return total_calls / number_of_microservices if number_of_microservices > 0 else 0

    def get_performance_metrics(self) -> Tuple[Dict[str, int], Dict[str, int], float]:
        """
        Get all performance metrics: Calls, Requests, and Avg. Calls.
        """
        calls_metrics = self.calculate_calls_metric()
        requests_metrics = self.calculate_requests_metric()
        avg_calls = self.calculate_average_calls()

        return calls_metrics, requests_metrics, avg_calls

    def calculate_story_points(self) -> Dict[str, int]:
        """
        Calculate the total story points (P) for each microservice.
        """
        story_points = {}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            story_points[ms_id] = sum(story.story_points for story in stories)
        return story_points

    def calculate_cg(self) -> Dict[str, int]:
        """
        Calculate Cg for each microservice.
        Cg = P * (Calls + Requests)
        """
        story_points = self.calculate_story_points()
        calls_metrics = self.calculate_calls_metric()
        requests_metrics = self.calculate_requests_metric()

        cg = {}
        for ms_id in self.microservice_manager.microservices:
            cg[ms_id] = story_points[ms_id] * (calls_metrics[ms_id] + requests_metrics[ms_id])
        return cg

    def calculate_pf(self) -> Dict[str, int]:
        pf = {}
        for ms_id in self.microservice_manager.microservices:
            pf[ms_id] = self._calculate_max_depth(ms_id, set())
        return pf

    def _calculate_max_depth(self, ms_id: str, visited: Set[str]) -> int:
        if ms_id in visited:
            return 0
        visited.add(ms_id)
        max_depth = 0
        stories = self.microservice_manager.get_microservice_stories(ms_id)
        for story in stories:
            for dep in story.dependencies:
                for other_ms_id, other_ms in self.microservice_manager.microservices.items():
                    if other_ms_id != ms_id and dep in other_ms.stories:
                        depth = 1 + self._calculate_max_depth(other_ms_id, visited.copy())
                        max_depth = max(max_depth, depth)
        return max_depth

    def calculate_cognitive_complexity(self) -> float:
        cg = self.calculate_cg()
        story_points = self.calculate_story_points()
        pf = self.calculate_pf()
        wsic_total = self.calculate_wsic_total()
        siy_total = self.calculate_siy_total()

        n = len(self.microservice_manager.microservices)
        max_story_points = max(story_points.values()) if story_points else 0

        cxt = (sum(cg.values()) / 2) * \
              (max_story_points / 2) * \
              (n / 2) * \
              (wsic_total / 2) * \
              (sum(pf.values()) / 2) * \
              ((siy_total + 2) / 2)

        return cxt

    def get_complexity_metrics(self) -> Tuple[Dict[str, int], Dict[str, int], Dict[str, int], float]:
        """
        Get all complexity metrics: Story Points (P), Cg, Pf, and Cognitive Complexity Points (CxT).
        """
        story_points = self.calculate_story_points()
        cg = self.calculate_cg()
        pf = self.calculate_pf()
        cxt = self.calculate_cognitive_complexity()

        return story_points, cg, pf, cxt

    def calculate_microservice_development_time(self) -> Dict[str, float]:
        """
        Calculate the estimated development time (ti) for each microservice.
        ti is the sum of the estimated development time of each user story in the microservice.
        """
        development_times = {}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            development_times[ms_id] = sum(story.estimated_time for story in stories)
        return development_times

    def calculate_msba_development_time(self) -> float:
        """
        Calculate the estimated development time (T) for the entire MSBA.
        T is the maximum development time among all microservices.
        """
        development_times = self.calculate_microservice_development_time()
        return max(development_times.values()) if development_times else 0

    def get_development_time_metrics(self) -> Tuple[Dict[str, float], float]:
        """
        Get all development time metrics:
        Microservice Development Time (ti) and MSBA Development Time (T).
        """
        microservice_times = self.calculate_microservice_development_time()
        msba_time = self.calculate_msba_development_time()
        return microservice_times, msba_time

    def calculate_semantic_similarity(self) -> Dict[str, float]:
        """
        Calculate the Semantic Similarity (SSi) for each microservice.
        """
        similarities = {}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            story_ids = list(microservice.stories)
            if story_ids:
                similarity = self.semantic_calculator.calculate_microservice_similarity(story_ids)
                similarities[ms_id] = similarity
            else:
                similarities[ms_id] = 0.0
        return similarities

    def calculate_total_semantic_similarity(self) -> float:
        similarities = self.calculate_semantic_similarity()
        return 100 * sum(similarities.values()) / len(similarities) if similarities else 0

    def get_semantic_similarity_metrics(self) -> Tuple[Dict[str, float], float]:
        """
        Get all semantic similarity metrics:
        Microservice Semantic Similarity (SSi) and MSBA Semantic Similarity (SsT).
        """
        microservice_similarities = self.calculate_semantic_similarity()
        total_similarity = self.calculate_total_semantic_similarity()
        return microservice_similarities, total_similarity

    def verify_semantic_similarity(self):
        """
        Verify the semantic similarity dictionary and print its contents and statistics.
        """
        print("Verifying Semantic Similarity Dictionary:")
        self.semantic_calculator.print_similarity_dict()
        print("\n")
        self.semantic_calculator.print_similarity_dict_stats()

    def print_user_stories(self):
        print("User Stories:")
        for ms_id, microservice in self.microservice_manager.microservices.items():
            print(f"\nMicroservice {ms_id}:")
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            for story in stories:
                print(f"  ID: {story.identifier}")
                print(f"  Title: {story.title}")
                print(f"  Description: {story.description}")
                print("  ---")





ERROR_MESSAGES = {
    'story_exists': {
        'en': "A story with the identifier {} already exists.",
        'es': "Ya existe una historia con el identificador {}."
    },
    'invalid_story_points': {
        'en': "Story points must be a positive integer.",
        'es': "Los story points deben ser un número entero positivo."
    },
    'invalid_estimated_time': {
        'en': "Estimated time must be a positive number.",
        'es': "El tiempo estimado debe ser un número positivo."
    },
    'story_not_found': {
        'en': "There is no story with the identifier {}.",
        'es': "No existe una historia con el identificador {}."
    },
    'missing_dependencies': {
        'en': "The following stories are referenced but do not exist: {}",
        'es': "Las siguientes historias son referenciadas pero no existen: {}"
    },
    'microservice_empty': {
        'en': "Microservice cannot be empty.",
        'es': "El microservicio no puede estar vacío."
    },
    'story_already_assigned': {
        'en': "The story {} is already assigned to another microservice.",
        'es': "La historia {} ya está asignada a otro microservicio."
    },
    'story_not_in_backlog': {
        'en': "The story {} is not in the backlog.",
        'es': "La historia {} no está en el backlog."
    }
}


def get_error_message(key: str, language: str) -> str:
    return ERROR_MESSAGES[key][language]


def add_language(language: str, messages: dict) -> None:
    for key in ERROR_MESSAGES:
        if key in messages:
            ERROR_MESSAGES[key][language] = messages[key]
        else:
            print(f"Warning: Message for key '{key}' not provided for language '{language}'")



import math


class FitnessFunctions:
    def __init__(self, metrics_manager):
        self.metrics_manager = metrics_manager

    def f1(self) -> float:
        cpt, _, wsict, cxt, sst = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + cxt ** 2 + wsict ** 2 + (100 - sst) ** 2) / 2

    def f2(self) -> float:
        cpt, _, wsict, _, sst = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + wsict ** 2 + (100 - sst) ** 2) / 2

    def f3(self) -> float:
        _, _, _, cxt, sst = self._get_metrics()
        return math.sqrt(cxt ** 2 + (100 - sst) ** 2) / 2

    def f4(self) -> float:
        cpt, coht, _, _, sst = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + coht ** 2 + (100 - sst) ** 2) / 2

    def f5(self) -> float:
        cpt, _, _, _, sst = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + (100 - sst) ** 2) / 2

    def f6(self) -> float:
        cpt, coht, wsict, _, sst = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + coht ** 2 + wsict ** 2 + (100 - sst) ** 2) / 2

    def f7(self) -> float:
        cpt, _, _, cxt, sst = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + cxt ** 2 + (100 - sst) ** 2) / 2

    def f8(self) -> float:
        cpt, coht, wsict, _, _ = self._get_metrics()
        return math.sqrt((10 * cpt) ** 2 + coht ** 2 + wsict ** 2) / 2

    def _get_metrics(self):
        cpt = self.metrics_manager.calculate_coupling_total()
        coht = self.metrics_manager.calculate_coh_total()
        wsict = self.metrics_manager.calculate_wsic_total()
        cxt = self.metrics_manager.calculate_cognitive_complexity()
        sst = self.metrics_manager.calculate_total_semantic_similarity()
        return cpt, coht, wsict, cxt, sst



from typing import List

import spacy


class SemanticSimilarityCalculator:
    def __init__(self, user_story_manager):
        self.nlp = spacy.load("en_core_web_md")
        self.similarity_dict = {}
        self.user_story_manager = user_story_manager

    def preprocess_text(self, text: str) -> str:
        doc = self.nlp(text)
        lemmas = [token.lemma_ for token in doc if token.pos_ in ["NOUN", "VERB", "ADJ"]]
        return " ".join(lemmas)

    def calculate_similarity(self, story_id1: str, story_id2: str) -> float:
        story1 = self.user_story_manager.get_story(story_id1)
        story2 = self.user_story_manager.get_story(story_id2)
        if story1 is None or story2 is None:
            return 0.0
        text1 = f"{story1.title} {story1.description}"
        text2 = f"{story2.title} {story2.description}"
        doc1 = self.nlp(self.preprocess_text(text1))
        doc2 = self.nlp(self.preprocess_text(text2))
        similarity = doc1.similarity(doc2)
        return similarity

    def calculate_microservice_similarity(self, story_ids: List[str]) -> float:
        total_similarity = 0.0
        comparisons = 0
        for i, id1 in enumerate(story_ids):
            for id2 in story_ids[i+1:]:
                similarity = self.calculate_similarity(id1, id2)
                total_similarity += similarity
                comparisons += 1
        return total_similarity / comparisons if comparisons > 0 else 0.0



import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import networkx as nx


class MicroserviceVisualizer:
    def __init__(self, microservice_manager):
        self.microservice_manager = microservice_manager

    def plot_microservices(self, base_color='lightblue', title=None):
        G = nx.DiGraph()

        # Create nodes and calculate sizes
        max_story_points = 1  # To avoid division by zero
        for ms_id, microservice in self.microservice_manager.microservices.items():
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            total_story_points = sum(story.story_points for story in stories)
            G.add_node(ms_id, story_points=total_story_points)
            max_story_points = max(max_story_points, total_story_points)

        # Add edges based on dependencies and calculate weights
        edge_weights = {}
        for ms_id, microservice in self.microservice_manager.microservices.items():
            stories = self.microservice_manager.get_microservice_stories(ms_id)
            for story in stories:
                for dep in story.dependencies:
                    for other_ms_id, other_microservice in self.microservice_manager.microservices.items():
                        if other_ms_id != ms_id:
                            if dep in other_microservice.stories:
                                edge = (other_ms_id, ms_id)
                                edge_weights[edge] = edge_weights.get(edge, 0) + 1

        # Add edges to the graph with calculated weights
        for edge, weight in edge_weights.items():
            G.add_edge(edge[0], edge[1], weight=weight)

        # Create figure and axes
        fig, ax = plt.subplots(figsize=(16, 12))

        # Use a layout that separates the nodes more
        pos = nx.spring_layout(G, k=1.5, iterations=50)

        # Draw nodes with color based on story points
        node_sizes = [G.nodes[node]['story_points'] * 30 for node in G.nodes()]
        node_colors = [self.adjust_color_intensity(base_color, G.nodes[node]['story_points'] / max_story_points) for
                       node in G.nodes()]
        nx.draw_networkx_nodes(G, pos, node_size=node_sizes, node_color=node_colors, ax=ax)

        # Draw edges
        edge_weights = [G[u][v]['weight'] for u, v in G.edges()]
        nx.draw_networkx_edges(G, pos, ax=ax, width=edge_weights, edge_color='gray',
                               arrows=True, arrowsize=20, arrowstyle='->', connectionstyle='arc3,rad=0.1')

        # Add node labels
        nx.draw_networkx_labels(G, pos, font_size=10, font_weight="bold", ax=ax)

        # Add edge labels (weight)
        edge_labels = nx.get_edge_attributes(G, 'weight')
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_size=8, ax=ax)

        if title:
            ax.set_title(title)
        else:
            ax.set_title("Microservices Dependency Graph")

        ax.axis('off')

        # Adjust graph limits
        ax.margins(0.2)  # Add margin around the graph

        # Adjust layout and display
        fig.tight_layout()
        plt.show()

    def adjust_color_intensity(self, base_color, intensity):
        """Adjust the intensity of the base color based on the number of story points"""
        rgb = mcolors.to_rgb(base_color)
        hsv = mcolors.rgb_to_hsv(rgb)
        # Adjust saturation and value (brightness) based on intensity
        hsv[1] = min(hsv[1] + 0.6 * intensity, 1.0)  # Increase saturation
        hsv[2] = max(hsv[2] - 0.6 * intensity, 0.2)  # Decrease brightness (darker)
        return mcolors.hsv_to_rgb(hsv)



from microservice_story_manager import (
    FitnessFunctions,
    GeneticMicroserviceAssigner,
    MicroserviceManager,
    MicroserviceMetricsManager,
    MicroserviceVisualizer,
    UserStoryManager,
)

user_story_manager = UserStoryManager()
user_story_manager.set_project_language('en')
user_story_manager.load_stories_from_csv('../examples/user_stories.csv')

user_story_manager.plot_stories()

microservice_manager = MicroserviceManager(user_story_manager)
microservice_manager.set_project_language('en')

# Create some microservices
for _ in range(8):
    microservice_manager.create_microservice()


# Define a simple fitness function (to be replaced with a more sophisticated one)
def simple_fitness(assignment):
    # This fitness function aims to balance the number of stories across microservices
    microservice_loads = [0] * len(microservice_manager.microservices)
    for ms_index in assignment:
        microservice_loads[ms_index] += 1
    return (1 / (max(microservice_loads) - min(microservice_loads) + 1),)


# Create and run the genetic algorithm
assigner = GeneticMicroserviceAssigner(user_story_manager, microservice_manager)
assigner.optimize_assignment(simple_fitness, population_size=100, generations=50)

# Validate and visualize the result
try:
    microservice_manager.validate_all_stories_assigned()
    print("All stories are assigned to microservices.")
except ValueError as e:
    print(f"Error: {str(e)}")

visualizer = MicroserviceVisualizer(microservice_manager)
visualizer.plot_microservices(base_color='lightgreen')


metrics_manager = MicroserviceMetricsManager(microservice_manager)

ais_total, ads_total, siy_total = metrics_manager.get_coupling_indicators()
print(f"AIS Total: {ais_total:.2f}")
print(f"ADS Total: {ads_total:.2f}")
print(f"SIY Total: {siy_total:.2f}")

cpt = metrics_manager.calculate_coupling_total()
print(f"Coupling Total (CpT): {cpt:.2f}")

coh_vector = metrics_manager.get_coh_vector()
print(f"Vector Coh: {coh_vector}")

lc_metrics, coh_metrics, coh_total = metrics_manager.get_cohesion_metrics()
print("Métricas de cohesión:")
print(f"LC: {lc_metrics}")
print(f"Coh: {coh_metrics}")
print(f"Coh Total: {coh_total:.4f}")



n, wsic_metrics, wsic_total = metrics_manager.get_granularity_metrics()
print("Métricas de granularidad:")
print(f"N: {n}")
print(f"WSIC: {wsic_metrics}")
print(f"WsicT: {wsic_total}")

calls_metrics, requests_metrics, avg_calls = metrics_manager.get_performance_metrics()
print("Métricas de rendimiento:")
print(f"Calls: {calls_metrics}")
print(f"Requests: {requests_metrics}")
print(f"Avg. Calls: {avg_calls:.2f}")

story_points, cg, pf, cxt = metrics_manager.get_complexity_metrics()

print("Métricas de complejidad:")

print(f"\nCognitive Complexity Points (CxT): {cxt:.2f}")


microservice_times, msba_time = metrics_manager.get_development_time_metrics()

print("Métricas de Tiempo de Desarrollo:")
print("\nTiempo de Desarrollo por Microservicio (ti):")
for ms_id, time in microservice_times.items():
    print(f"  Microservice {ms_id}: {time:.2f} horas")

print(f"\nTiempo de Desarrollo Estimado de MSBA (T): {msba_time:.2f} horas")

metrics_manager.verify_semantic_similarity()

print("Printing User Stories:")
metrics_manager.print_user_stories()

print("\nVerifying Semantic Similarity:")
metrics_manager.verify_semantic_similarity()

microservice_similarities, total_similarity = metrics_manager.get_semantic_similarity_metrics()
print("Métricas de Similitud Semántica:")
print("\nSimilitud Semántica por Microservicio (SSi):")
for ms_id, similarity in microservice_similarities.items():
    print(f"  Microservice {ms_id}: {similarity:.2f}")

print(f"\nSimilitud Semántica Total de MSBA (SsT): {total_similarity:.2f}")


fitness1 = FitnessFunctions(metrics_manager).f1()
fitness2 = FitnessFunctions(metrics_manager).f2()
print(fitness1, fitness2)



from .user_story_manager import UserStory, UserStoryManager
from .microservice_manager import MicroserviceManager
from .semantic_similarity import SemanticSimilarityCalculator
from .fitness_functions import FitnessFunctions
from .metrics_visualizer import MetricsVisualizer
from .microservice_visualizer import MicroserviceVisualizer

from .microservice_metrics import MicroserviceMetricsManager

from .genetic_microservice_assigner import GeneticMicroserviceAssigner

__all__ = [
    "UserStory", "UserStoryManager", "MicroserviceManager", "MicroserviceMetricsManager",
    "SemanticSimilarityCalculator", "FitnessFunctions", "MetricsVisualizer",
    "MicroserviceVisualizer", "GeneticMicroserviceAssigner"
]



# metrics_visualizer.py
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np


class MetricsVisualizer:
    @staticmethod
    def compare_metrics(metrics1: Dict, metrics2: Dict, name1: str, name2: str):
        metrics = list(metrics1.keys())
        values1 = list(metrics1.values())
        values2 = list(metrics2.values())

        x = np.arange(len(metrics))
        width = 0.35

        fig, ax = plt.subplots(figsize=(12, 6))
        rects1 = ax.bar(x - width / 2, values1, width, label=name1)
        rects2 = ax.bar(x + width / 2, values2, width, label=name2)

        ax.set_ylabel('Metric Values')
        ax.set_title('Comparison of Metrics between Two Fitness Functions')
        ax.set_xticks(x)
        ax.set_xticklabels(metrics)
        ax.legend()

        def autolabel(rects):
            for rect in rects:
                height = rect.get_height()
                ax.annotate(f'{height:.2f}',
                            xy=(rect.get_x() + rect.get_width() / 2, height),
                            xytext=(0, 3),
                            textcoords="offset points",
                            ha='center', va='bottom')

        autolabel(rects1)
        autolabel(rects2)

        fig.tight_layout()
        plt.show()

    @staticmethod
    def compare_fitness_values(fitness_values: List[float], function_names: List[str]):
        x = np.arange(len(function_names))

        fig, ax = plt.subplots(figsize=(10, 6))
        bars = ax.bar(x, fitness_values, width=0.6)

        ax.set_ylabel('Fitness Value')
        ax.set_title('Comparison of Fitness Values')
        ax.set_xticks(x)
        ax.set_xticklabels(function_names)

        def autolabel(rects):
            for rect in rects:
                height = rect.get_height()
                ax.annotate(f'{height:.2f}',
                            xy=(rect.get_x() + rect.get_width() / 2, height),
                            xytext=(0, 3),
                            textcoords="offset points",
                            ha='center', va='bottom')

        autolabel(bars)

        fig.tight_layout()
        plt.show()



import random
from typing import Callable, List, Tuple

from deap import algorithms, base, creator, tools

from microservice_story_manager import MicroserviceManager, UserStoryManager


class GeneticMicroserviceAssigner:
    def __init__(self, user_story_manager: UserStoryManager, microservice_manager: MicroserviceManager):
        self.user_story_manager = user_story_manager
        self.microservice_manager = microservice_manager
        self.stories = list(user_story_manager.get_all_stories().keys())
        self.microservices = list(microservice_manager.microservices.keys())

    def setup_genetic_algorithm(self, fitness_func: Callable[[List[int]], Tuple[float,]]):
        creator.create("FitnessMax", base.Fitness, weights=(1.0,))
        creator.create("Individual", list, fitness=creator.FitnessMax)

        self.toolbox = base.Toolbox()
        self.toolbox.register("attr_int", random.randint, 0, len(self.microservices) - 1)
        self.toolbox.register("individual", tools.initRepeat, creator.Individual, self.toolbox.attr_int,
                              n=len(self.stories))
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)

        self.toolbox.register("evaluate", fitness_func)
        self.toolbox.register("mate", tools.cxTwoPoint)
        self.toolbox.register("mutate", tools.mutUniformInt, low=0, up=len(self.microservices) - 1, indpb=0.1)
        self.toolbox.register("select", tools.selTournament, tournsize=3)

    def run_genetic_algorithm(self, population_size: int = 100, generations: int = 50) -> List[int]:
        population = self.toolbox.population(n=population_size)
        algorithms.eaSimple(population, self.toolbox, cxpb=0.7, mutpb=0.2, ngen=generations, verbose=False)

        best_individual = tools.selBest(population, k=1)[0]
        return best_individual

    def assign_stories_to_microservices(self, assignment: List[int]) -> None:
        # Clear existing assignments
        for ms in self.microservice_manager.microservices.values():
            ms.stories.clear()

        # Assign stories based on the genetic algorithm result
        for story_id, ms_index in zip(self.stories, assignment):
            ms_id = self.microservices[ms_index]
            self.microservice_manager.add_story_to_microservice(ms_id, story_id)

    def optimize_assignment(self, fitness_func: Callable[[List[int]], Tuple[float,]], population_size: int = 100,
                            generations: int = 50) -> None:
        self.setup_genetic_algorithm(fitness_func)
        best_assignment = self.run_genetic_algorithm(population_size, generations)
        self.assign_stories_to_microservices(best_assignment)



import random
from typing import Callable, Dict, List, Tuple

import numpy as np  # Add this line
from deap import algorithms, base, creator, tools

from microservice_story_manager import MicroserviceManager, MicroserviceMetricsManager


class GeneticMicroserviceAssigner:
    def __init__(self, user_story_manager, microservice_manager, fitness_function: Callable, minimize: bool = True):
        self.user_story_manager = user_story_manager
        self.microservice_manager = microservice_manager
        self.fitness_function = fitness_function
        self.minimize = minimize

        from microservice_story_manager import MicroserviceMetricsManager
        self.metrics_manager = MicroserviceMetricsManager(microservice_manager)

        self.n_stories = len(self.user_story_manager.get_all_stories())
        self.n_microservices = len(self.microservice_manager.microservices)

    def create_individual(self) -> List[int]:
        """Create a random individual (chromosome)"""
        return [random.randint(0, self.n_microservices - 1) for _ in range(self.n_stories)]

    def setup_genetic_algorithm(self):
        if self.minimize:
            creator.create("Fitness", base.Fitness, weights=(-1.0,))
        else:
            creator.create("Fitness", base.Fitness, weights=(1.0,))

        creator.create("Individual", list, fitness=creator.Fitness)

        toolbox = base.Toolbox()
        toolbox.register("individual", tools.initIterate, creator.Individual, self.create_individual)
        toolbox.register("population", tools.initRepeat, list, toolbox.individual)

        def evaluate(individual):
            self.assign_stories_to_microservices(individual)
            return (self.fitness_function(),)  # Note: removed self.metrics_manager

        toolbox.register("evaluate", evaluate)
        toolbox.register("mate", tools.cxTwoPoint)
        toolbox.register("mutate", tools.mutUniformInt, low=0, up=self.n_microservices - 1, indpb=0.1)
        toolbox.register("select", tools.selTournament, tournsize=3)

        return toolbox

    def assign_stories_to_microservices(self, individual: List[int]) -> None:
        """Assign stories to microservices based on the individual"""
        # Clear existing assignments
        for ms in self.microservice_manager.microservices.values():
            ms.stories.clear()

        # Assign stories based on the individual
        stories = list(self.user_story_manager.get_all_stories().keys())
        for story_idx, ms_idx in enumerate(individual):
            ms_id = list(self.microservice_manager.microservices.keys())[ms_idx]
            self.microservice_manager.add_story_to_microservice(ms_id, stories[story_idx])

        # Print final assignments
        """for ms_id, ms in self.microservice_manager.microservices.items():
            print(f"Microservice {ms_id} has stories: {ms.stories}")"""

    def run_genetic_algorithm(self, toolbox, generations=50, population_size=100):
        """Run the genetic algorithm"""
        population = toolbox.population(n=population_size)
        hof = tools.HallOfFame(1)
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean)
        stats.register("min", np.min)
        stats.register("max", np.max)

        population, logbook = algorithms.eaSimple(population, toolbox, cxpb=0.7, mutpb=0.2,
                                                  ngen=generations, stats=stats, halloffame=hof, verbose=True)

        return hof[0], logbook

    def optimize_assignments(self, generations=50, population_size=100) -> Tuple[object, float, Dict, Dict[str, List[str]]]:
        from microservice_story_manager import MetricsVisualizer
        print("\nOptimizing using provided fitness function")
        toolbox = self.setup_genetic_algorithm()
        best_individual, logbook = self.run_genetic_algorithm(toolbox, generations, population_size)

        optimized_microservice_manager = self.create_optimized_microservice_manager(best_individual)

        # Calculate detailed metrics
        detailed_metrics = self.calculate_detailed_metrics(optimized_microservice_manager)

        # Get story assignments
        story_assignments = self.get_story_assignments(optimized_microservice_manager)

        return optimized_microservice_manager, best_individual.fitness.values[0], detailed_metrics, story_assignments

    def get_story_assignments(self, microservice_manager: MicroserviceManager) -> Dict[str, List[str]]:
        assignments = {}
        for ms_id, microservice in microservice_manager.microservices.items():
            stories = microservice_manager.get_microservice_stories(ms_id)
            assignments[ms_id] = [f"{story.title}" for story in stories]
        return assignments

    def calculate_detailed_metrics(self, microservice_manager: MicroserviceManager) -> Dict:
        metrics = MicroserviceMetricsManager(microservice_manager)
        return {
            "coupling": metrics.calculate_coupling_total(),
            "cohesion": metrics.calculate_coh_total(),
            "granularity": metrics.calculate_wsic_total(),
            "complexity": metrics.calculate_cognitive_complexity(),
            "semantic_similarity": metrics.calculate_total_semantic_similarity()
        }

    def create_optimized_microservice_manager(self, individual: List[int]) -> MicroserviceManager:
        """Create a new MicroserviceManager with the optimized assignments"""
        optimized_manager = MicroserviceManager(self.user_story_manager)

        # Create the same number of microservices as in the original manager
        for _ in range(self.n_microservices):
            optimized_manager.create_microservice()

        # Assign stories based on the individual
        stories = list(self.user_story_manager.get_all_stories().keys())
        for story_idx, ms_idx in enumerate(individual):
            ms_id = list(optimized_manager.microservices.keys())[ms_idx]
            optimized_manager.add_story_to_microservice(ms_id, stories[story_idx])

        return optimized_manager

    def get_microservice_assignments(self, chromosome: List[int]) -> Dict[str, List[str]]:
        """Get microservice assignments from a chromosome"""
        assignments = {ms_id: [] for ms_id in self.microservice_manager.microservices}
        stories = list(self.user_story_manager.get_all_stories().keys())
        for story_idx, ms_idx in enumerate(chromosome):
            ms_id = list(self.microservice_manager.microservices.keys())[ms_idx]
            assignments[ms_id].append(stories[story_idx])
        return assignments

    def print_results(self, results: Dict[str, Tuple[List[int], float]]) -> None:
        """Print optimization results"""
        for fitness_name, (best_chromosome, fitness_value) in results.items():
            print(f"\nBest assignment for fitness function {fitness_name}:")
            assignments = self.get_microservice_assignments(best_chromosome)
            for ms_id, stories in assignments.items():
                print(f"  Microservice {ms_id}: {stories}")

            print(f"  Fitness value: {fitness_value}")

            # Calculate and display metrics for this assignment
            self.assign_stories_to_microservices(best_chromosome)
            cpt = self.metrics_manager.calculate_coupling_total()
            coht = self.metrics_manager.calculate_coh_total()
            wsict = self.metrics_manager.calculate_wsic_total()
            cxt = self.metrics_manager.calculate_cognitive_complexity()
            sst = self.metrics_manager.calculate_total_semantic_similarity()

            print("  Metrics:")
            print(f"    Coupling (CpT): {cpt}")
            print(f"    Cohesion (CohT): {coht}")
            print(f"    Granularity (WsicT): {wsict}")
            print(f"    Complexity (CxT): {cxt}")
            print(f"    Semantic Similarity (SsT): {sst}")


"""
user_story_manager = UserStoryManager()
user_story_manager.set_project_language('en')
user_story_manager.load_stories_from_csv('../examples/user_stories.csv')

microservice_manager = MicroserviceManager(user_story_manager)
microservice_manager.set_project_language('en')

# Create initial microservices (adjust the number as needed)
for _ in range(8):
    microservice_manager.create_microservice()

    # Create and run the genetic algorithm
# Create FitnessFunctions instance
fitness_functions = FitnessFunctions(MicroserviceMetricsManager(microservice_manager))


# Define the two fitness functions to compare
functions_to_compare = [
    (fitness_functions.f1, "f1", True),
    (fitness_functions.f2, "f2", True)
]

results = []

for func, name, minimize in functions_to_compare:
    assigner = GeneticMicroserviceAssigner(user_story_manager, microservice_manager, func, minimize)
    optimized_manager, fitness_value, detailed_metrics, story_assignments = assigner.optimize_assignments(generations=5,
                                                                                                          population_size=10)
    results.append((name, optimized_manager, fitness_value, detailed_metrics, story_assignments))

    print(f"\nResultados para la función de fitness {name}:")
    print(f"Mejor fitness: {fitness_value}")
    print(f"Métricas detalladas: {detailed_metrics}")
    print("\nAsignaciones de historias:")
    for ms_id, stories in story_assignments.items():
        print(f"\nMicroservicio {ms_id}:")
        for story in stories:
            print(f"  - {story}")

# Compare metrics
MetricsVisualizer.compare_metrics(
    results[0][3], results[1][3],
    results[0][0], results[1][0]
)

# Compare fitness values
MetricsVisualizer.compare_fitness_values(
    [result[2] for result in results],
    [result[0] for result in results]
)

# Visualize the optimized microservices
for name, manager, _, _, _ in results:
    visualizer = MicroserviceVisualizer(manager)
    visualizer.plot_microservices(title=f"Microservicios optimizados usando {name}")
"""
